=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed March 2010 by Giovanni Di Trapani

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   kicad_class.rb
# Version	:	1.5
# Type		:   Sketchup Tool
# Description	:   Set of classes used to draw kicad elements on Skecthup
# Usage		:   Used by kicad_parser.rb
# Original Date	:   18 March 2010
# Last Change	:   13 Feb 2011
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup'


class Kicad_base

	COPPER_BOTTOM=0; COPPER_LAYER1=1; COPPER_LAYER2=2; COPPER_LAYER3=3
	COPPER_LAYER4=4; COPPER_LAYER5=5; COPPER_LAYER6=6; COPPER_LAYER7=7
	COPPER_LAYER8=8; COPPER_LAYER9=9; COPPER_LAYER10=10; COPPER_LAYER11=11
	COPPER_LAYER12=12; COPPER_LAYER13=13; COPPER_LAYER14=14; COPPER_TOP=15
	BOTTOM_SIDE_ADHESIVE= 16; TOP_SIDE_ADHESIVE = 17; BOTTOM_SOLDER_PASTE = 18; TOP_SOLDER_PASTE = 19
	BOTTOM_SILK_SCREEN = 20; TOP_SILK_SCREEN = 21; BOTTOM_SOLDER_MASK = 22; TOP_SOLDER_MASK = 23
	DRAWINGS = 24; COMMENT = 25; ECO1 = 26; ECO2 = 27; EDGE = 28
	
	SHAPE3D_LAYER=29 # shape3d special layer

	attr_accessor :unit, :z, :layer_name
	
	@@unit=0.0001
	@@z=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
	@@layer_name=["layer0", "layer1", "layer2", "layer3", "layer4", "layer5", "layer6", "layer7", 
				"layer8","layer9", "layer10", "layer11", "layer12", "layer13", "layer14", "layer15",
				"bottom side adhesive", "top side adhesive", "bottom solder paste", "top solder paste", 
				"bottom silk screen", "top silk screen", "bottom solder mask", "top solder mask",
				"drawings", "comment", "ECO1", "ECO2", "edge", "SHAPE3D"]

	def mirror_x(element,origin)
		vector=Geom::Vector3d.new(1,0,0)
		angle=180
		mirror=Geom::Transformation.rotation origin, vector, angle.degrees
		element.transform! mirror
	end
	
	def mirror_y(element,origin)
		vector=Geom::Vector3d.new(0,1,0)
		angle=180
		mirror=Geom::Transformation.rotation origin, vector, angle.degrees
		element.transform! mirror
	end

	def mirror_z(element,origin)
		vector=Geom::Vector3d.new(0,0,1)
		angle=180
		mirror=Geom::Transformation.rotation origin, vector, angle.degrees
		element.transform! mirror
	end

	def rotate(element,origin,angle)
		vector=Geom::Vector3d.new(0,0,1)
		rotate=Geom::Transformation.rotation origin, vector, angle.degrees
		element.transform! rotate
	end
	
	def draw_rect(group,origin,x_size,y_size,orientation)
		points=[]
		edges=[]
		x_delta = x_size/2*@@unit
		y_delta = y_size/2*@@unit
		points<<Geom::Point3d.new(origin.x-x_delta,origin.y-y_delta,origin.z)
		points<<Geom::Point3d.new(origin.x+x_delta,origin.y-y_delta,origin.z)
		points<<Geom::Point3d.new(origin.x+x_delta,origin.y+y_delta,origin.z)
		points<<Geom::Point3d.new(origin.x-x_delta,origin.y+y_delta,origin.z)
		points.each do |point|
			if(orientation!=0)
				rotate(point,origin,orientation)
			end
		end
		edges<<(group.entities.add_line points[0],points[1])
		edges<<(group.entities.add_line points[1],points[2])
		edges<<(group.entities.add_line points[2],points[3])
		edges<<(group.entities.add_line points[3],points[0])		
		edges
	end
	
	def draw_circle(group,center,diameter)
		vector=Geom::Vector3d.new(0,0,1)
		circle=(group.entities.add_circle center,vector,diameter/2*@@unit)
		circle
	end
	
	def draw_oblong(group,origin,x_size,y_size,orientation)
		oblong=[]
		arc_radius=(x_size>y_size) ? y_size/2*@@unit : x_size/2*@@unit
		xaxis=Geom::Vector3d.new(1,0,0)
		normal=Geom::Vector3d.new(0,0,1)
		
		x_lenght=(x_size>y_size) ? (x_size-y_size)*@@unit : 0
		y_lenght=(x_size>y_size) ? 0 : (y_size-x_size)*@@unit

		center_1=Geom::Point3d.new(origin.x+x_lenght/2,origin.y+y_lenght/2,origin.z)
		center_2=Geom::Point3d.new(origin.x-x_lenght/2,origin.y-y_lenght/2,origin.z)
		if(orientation!=0)
			rotate(center_1,origin,orientation)
			rotate(center_2,origin,orientation)
		end
		
		start_angle=(x_size>y_size) ? -90 : 0
		end_angle=(x_size>y_size) ? 90 : 180
		start_angle+=orientation
		end_angle+=orientation
		
		arc_1=(group.entities.add_arc center_1, xaxis, normal, arc_radius, start_angle.degrees, end_angle.degrees)
		p_00=arc_1.first.vertices[0].position
		p_01=arc_1.last.vertices[1].position

		start_angle=(x_size>y_size) ? 90 : 180
		end_angle=(x_size>y_size) ? 270 : 360
		start_angle+=orientation
		end_angle+=orientation

		arc_2=(group.entities.add_arc center_2, xaxis, normal, arc_radius, start_angle.degrees, end_angle.degrees)
		p_10=arc_2.first.vertices[0].position
		p_11=arc_2.last.vertices[1].position
		line_1=group.entities.add_line p_00, p_11
		line_2=group.entities.add_line p_01, p_10
		
		oblong=[line_2,line_1]+arc_2+arc_1
		oblong
	end

	def draw_arc(group,center,e_point,angle)
		radius=center.distance e_point
		vector=center.vector_to(e_point)
		vector.normalize!
		normal=Geom::Vector3d.new(0,0,-1)
		s_angle=0
		e_angle=s_angle+angle
		arc_center=Geom::Point3d.new(center.x*@@unit,center.y*@@unit,center.z)
		arc=group.entities.add_arc arc_center,vector,normal,radius*@@unit,s_angle.degrees,e_angle.degrees
		arc
	end
	
	def draw_text(group,origin,x_size,y_size,o_angle,mirror,text)
		
		x_displace=text.length*x_size/2
		y_displace=y_size/2

	# the following condition has been added in order to avoid a crash
		if(y_size<200) then y_size=200
		end
		
		pos=Geom::Point3d.new(origin.x-x_displace*@@unit,origin.y-y_displace*@@unit,origin.z)
				
		t_group=group.entities.add_group
		t_group.name="REFERENCE"

		t_group.entities.add_3d_text(text, TextAlignCenter, "Courier", false, false, y_size*@@unit, 0, 0, true, 0)
		t_group.move! pos
		
		if(mirror==1)
			mirror_y(t_group,origin)
		end
		
		if(o_angle!=0)
			rotate(t_group,origin,o_angle*0.1)
		end
		faces=inventory_faces(t_group)
		faces.each do |face|
			face.material=$kicad_settings.silk_color
			face.back_material=$kicad_settings.silk_color
		end

		t_group
	
	end

	def inventory_faces(group)
		faces=[]
		group.entities.each {|ent|
		faces.push(ent) if ent.is_a? Sketchup::Face }
		return faces
	end

end

class Kicad_Field < Kicad_base

	attr_accessor :Xoff, :Yoff, :Xsize, :Ysize, :Orientation, :Text
	
	def initialize
		@Xoff=0
		@Yoff=0
		@Xsize=0
		@Ysize=0
		@Orientation=0
		@Text=""
	end

	
end

class Kicad_report

	attr_accessor :pcb_rev, :pcb_title, :pcb_date
	attr_reader :pad_elaps, :drawing_elaps, :drill_elaps, :shape3d_elaps; :total_elaps
	attr_accessor :total_pads, :total_components, :total_shape3d, :missing_shape3d, :missing_drills
	
	def initialize
		@total_pads=0
		@total_components=0
		@total_shape3d=0
		@missing_drills=0
		@missing_shape3d=0
		@pad_elaps=0
		@drawing_elaps=0
		@drill_elaps=0
		@shape3d_elaps=0
		@total_elaps=0
		@pcb_rev=""
		@pcb_title=""
		@pcb_date=""
	end
	
	def inc_total_components
		@total_components+=1
	end

	def inc_total_shape3d
		@total_shape3d+=1
	end
	
	def inc_total_pads
		@total_pads+=1
	end

	def inc_missing_shape3d
		@missing_shape3d+=1
	end

	def inc_missing_drills
		@missing_drills+=1
	end

	def update_pad_elaps(value)
		@pad_elaps+=value
	end

	def update_drawing_elaps(value)
		@drawing_elaps+=value
	end

	def update_drill_elaps(value)
		@drill_elaps+=value
	end

	def update_total_elaps(value)
		@total_elaps+=value
	end
	
	def show

		report=""

		report+="KICAD BOARD: "
		if(@pcb_title!="")
			report+="#{@pcb_title}\n"
		else report+="No Title\n"
		end
		
		if(@pcb_rev!="") then report+="Revision: #{@pcb_rev}\n"
		end

		if(@pcb_date!="") then report+="Date: #{@pcb_date}\n"
		end
		
		report+="-------------------------------------\n"

		report+="Total componets: #{@total_components}\n"
		report+="Total pads: #{@total_pads}\n"
		report+="Total shape3d: #{@total_shape3d}\n"

		report+="-------------------------------------\n"

		report+="Missing shape3d: #{@missing_shape3d}\n"
		report+="Missing drills: #{@missing_drills}\n"

		report+="-------------------------------------\n"

		pad_elaps=sprintf("%.1f",@pad_elaps)
		report+="Time for pads: #{pad_elaps} sec\n"

		drawing_elaps=sprintf("%.1f",@drawing_elaps)
		report+="Time for drawings: #{drawing_elaps} sec\n"
		
		drill_elaps=sprintf("%.1f",@drill_elaps)
		report+="Time for drilling: #{drill_elaps} sec\n"

		report+="-------------------------------------\n"
		total_elaps=sprintf("%.1f",@total_elaps)
		report+="Total time: #{total_elaps} sec\n"

		
		UI.messagebox(report)
	end
end

class Kicad_settings

	attr_reader :sys_location, :user_path, :thickness, :pcb_color, :pad_color, :silk_color, :reference
	attr_reader :draw_silk, :draw_pad, :draw_shape3d, :drill_pcb, :autoplace, :multilayer
	attr_reader :prompts, :defaults, :list
	
	THICKNESS=0
	PCB_COLOR=1
	PAD_COLOR=2
	SILK_COLOR=3
	REFERENCE=4
	DRAW_SILK=5
	DRAW_PAD=6
	DRAW_SHAPE3D=7
	DRILL_PCB=8
	AUTOPLACE=9
	MULTILAYERS=10
	
	def initialize
		@sys_location=1
		@user_path=nil
		@thickness=0.039
		@pcb_color="Green"
		@pad_color="Gold"
		@silk_color="White"
		@reference=0
		@draw_silk=1
		@draw_pad=1
		@draw_shape3d=1
		@drill_pcb=1
		@autoplace=1
		@multilayer=0
		
		@prompts=['Thickness','PCB color', 'PAD color','SILK color','Reference','SILK','PAD','SHAPE3D','DRILL PCB','AUTOPLACE','MULTILAYER']
		@defaults=['1mm','','','','','YES','YES','YES','YES','YES','NO']
		@list=["","Green|Red|Black|Sand","Gold|Silver","White|Snow|Yellow|Goldenrod|Silver","T0|T1","YES|NO","YES|NO","YES|NO","YES|NO","YES|NO","YES|NO"]
	end
	
	def import(settings)

		#set the component 3d location :TBD
				
		#Acquire the thickness of the PCB
		
		value=settings[THICKNESS].scan(/\d+\.?\d*/)
		unit=settings[THICKNESS].scan(/[a-zA-Z]+/)
		case unit[0]
		when "mm"
			@thickness=value[0].to_f.mm
		when "\""
			@thickness=value[0].to_f
		else
			@thickness=0.039
		end
		
		rgb={"Green" => "#107D58", "Red" => "#E12319", "Black" => "#0A0907", "Sand" =>"#9F8650"}
		
		#Set the PCB color
		
		@pcb_color=rgb[settings[PCB_COLOR]]
		
		#Set the PAD color
		
		@pad_color=settings[PAD_COLOR]

		#Set the SILK color
		
		@silk_color=settings[SILK_COLOR]
		
		#set which reference has to be drawn
		
		if(settings[REFERENCE]=="T0")
			@reference=0
		else
			@reference=1
		end
				
		#Set a boolean condition for drawing silk screen

		if(settings[DRAW_SILK]=="YES")
			@draw_silk=1
		else
			@draw_silk=0 
		end

		
		#Set a boolean condition for drawing pads
		if(settings[DRAW_PAD]=="YES")
			@draw_pad=1
		else
			@draw_pad=0 
		end
		
		#Set a boolean condition for drawing shapes
		if(settings[DRAW_SHAPE3D]=="YES")
			@draw_shape3d=1
		else
			@draw_shape3d=0 
		end
		
		if(settings[DRILL_PCB]=="YES")
			@drill_pcb=1
		else
			@drill_pcb=0
		end
		
		if(settings[AUTOPLACE]=="YES")
			@autoplace=1
		else
			@autoplace=0
		end
		if(settings[MULTILAYER]=="YES")
			@multilayer=1
		else
			@multilayer=0
		end
	end
	
end

class Drawsegment < Kicad_base

	attr_accessor :Xstart, :Ystart, :Xend, :Yend, :Xcentre, :Ycentre, :angle
	attr_accessor :shape, :layer

	LINE=0
	CIRCLE=3
	ARC=2

	def initialize
		@Xstart=0
		@Ystart=0
		@Xend=0
		@Yend=0
		@Xcentre=0
		@Ycentre=0
		@shape=0
		@angle=0
		@layer=0
		
	end
	
	def draw(group)
		
	#select model and active layer
		kicad_model = Sketchup.active_model
		kicad_layers = Sketchup.active_model.layers
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[@layer]
		end
		edge_array=[]
		case @shape
			when LINE
				start_line=Geom::Point3d.new(@Xstart*@@unit,@Ystart*@@unit,@@z[EDGE])
				end_line=Geom::Point3d.new(@Xend*@@unit,@Yend*@@unit,@@z[EDGE])			
				edge_array<<(group.entities.add_line start_line,end_line)

			when CIRCLE
				center=Geom::Point3d.new(@Xcentre,@Ycentre,@@z[EDGE])
				e_point=Geom::Point3d.new(@Xend,@Yend,@@z[EDGE])
				radius=center.distance e_point
				edge_array+=draw_circle(group,center,radius/2)
			when ARC
				center=Geom::Point3d.new(@Xcentre,@Ycentre,@@z[EDGE])
				e_point=Geom::Point3d.new(@Xend,@Yend,@@z[EDGE])
				edge_array+=draw_arc(group,center,e_point,@angle*0.1)
		end
		edge_array
	end

end

class Kicad_text < Kicad_base

	attr_accessor :Xstart, :Ystart, :Xsize, :Ysize ,:rotation
	attr_accessor :layer, :normal
	attr_accessor :text
	
	def initialize
	
		@Xstart=0
		@Ystart=0
		@Xsize=0
		@Ysize=0
		@rotation=0
		@layer=0
		@normal=0
		@text=""
	end
	
	def draw(group)
		
		kicad_layers = Sketchup.active_model.layers
		kicad_model = Sketchup.active_model
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[@layer]
		end

		x_displace=@text.length/2*@Xsize
		y_displace=@Ysize/2
		
		pos=Geom::Point3d.new((@Xstart-x_displace)*@@unit,(-y_displace-@Ystart)*@@unit,@@z[@layer])

		#Prepare transformation for rotation
		
		vector=Geom::Vector3d.new(0,0,1)
		rotation=Geom::Transformation.rotation pos, vector,@rotation.degrees*0.1
		
		if(@normal==1)
			mirror_angle=0
		else
			mirror_angle=180
		end
		
		vector=Geom::Vector3d.new(0,1,0)
		mirror=Geom::Transformation.rotation pos, vector,mirror_angle.degrees

		group.entities.add_3d_text(@text, 0, "Courier", false, false, @Ysize*@@unit, 0, 0, true, 0)
		group.move! pos
				
		group.transform! mirror

		center=Geom::Point3d.new(@Xstart*@@unit,(-y_displace-@Ystart)*@@unit,@@z[@layer])
		
		if(rotation!=0) then rotate(group,center,@rotation*0.1)
		end
		faces=inventory_faces(group)
		faces.each do |face|
			face.material=$kicad_settings.silk_color
			face.back_material=$kicad_settings.silk_color
		end
	
	end
	
	def inventory_faces(group)
		faces=[]
		group.entities.each {|ent|
		faces.push(ent) if ent.is_a? Sketchup::Face }
		return faces
	end
	
end

class Pad < Kicad_base

	attr_accessor :name, :shape, :type, :Xsize, :Ysize, :Xdelta, :Ydelta, :Orientation
	attr_accessor :layer_mask, :Xpos, :Ypos, :drill
	
	C=0; R=1; O=2; T=3

	STD=0; SMD=1; CONN=2; HOLE=3; MECA=4; UKNOWN=5
	
	TOP=0x8000; BOTTOM=0x1
	
	def initialize
		@Xpos=0
		@Ypos=0
		@layer_mask=0
		@Xsize=0
		@Ysize=0
		@Xdelta=0
		@Ydelta=0
		@Orientation=0
		@name=""
		@shape=C
		@type=STD
		@drill=nil
	end

	def draw(group,m_point,r_angle,comp_layer)
	
		kicad_layers = Sketchup.active_model.layers
		kicad_model = Sketchup.active_model
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[COPPER_TOP]
		end
		
		
		
		n_pos=Geom::Point3d.new(@Xpos*@@unit+m_point.x,@Ypos*@@unit+m_point.y,m_point.z)
		rotate(n_pos,m_point,r_angle*0.1)
		
		#Prepare transformation for orientation
		
		o_angle=@Orientation*0.1

	
		face_drill=nil


		#Draw the inner face for the drilling
		if(@drill!=nil)
			case @drill.type
				when Kicad_drill::CIRCLE
					center=@drill.center(n_pos)
					if(@drill.Xdrill>=0.020)
						drill_circle=draw_circle(group,center,@drill.Xdrill)
						face_drill=group.entities.add_face(drill_circle)
					end
				when Kicad_drill::OBLONG
					center=@drill.center(n_pos)
					if((@drill.Xdrill>=0.020) && (@drill.Ydrill>=0.020))
						drill_oblong=draw_oblong(group,center,@drill.Xdrill,@drill.Ydrill,o_angle)
						face_drill=group.entities.add_face(drill_oblong)
					end
			end
		end
		
		case @shape
			when C
				center=n_pos
				vector=Geom::Vector3d.new(0,0,1)
				circle=draw_circle(group,center,@Xsize)
				face=group.entities.add_face(circle)
			when R
				rect=draw_rect(group,n_pos,@Xsize,@Ysize,o_angle)
				face=group.entities.add_face(rect)
			when O
				oblong=draw_oblong(group,n_pos,@Xsize,@Ysize,o_angle)
				face=group.entities.add_face(oblong)
								
			when T
		end
		if(face)
			face.material=$kicad_settings.pad_color 
			face.back_material=$kicad_settings.pad_color
		end
		if(face_drill!=nil) then face_drill.erase!
		end
	end

	
end

class Kicad_drill < Kicad_base

	attr_accessor :Xoff, :Yoff, :Xdrill, :Ydrill, :type

	CIRCLE=0
	OBLONG=1
	UKNOWN=2
	
	def initialize
		@Xoff=0
		@Yoff=0
		@Xdrill=0
		@Ydrill=0
		@type=UKNOWN
	
	end
	
	def center(p_pos)
		center=Geom::Point3d.new(p_pos.x+@Xoff*@@unit,p_pos.y+@Yoff*@@unit,p_pos.z)
		return center
	end
		
end

class Shape3D < Kicad_base

	attr_accessor :filename, :id

	def initialize
		@filename=""
		@id=0
	end

	def place(group,m_point,r_angle,layer)
	
		kicad_model = Sketchup.active_model
		kicad_shape3d = kicad_model.definitions

		pos=Geom::Transformation.new m_point
		
		#Prepare transformation for mirroring
		vector=Geom::Vector3d.new(1,0,0)
		if(layer==COPPER_TOP) then angle=0
		else angle=180
		end
		
		mirror=Geom::Transformation.rotation m_point, vector, angle.degrees
		
		#Prepare tranformation for rotation
		
		vector.z=1
		vector.x=0
		
		rotate=Geom::Transformation.rotation m_point, vector, r_angle.degrees*0.1
		
#		shape=kicad_model.entities.add_instance(kicad_shape3d[@id],pos)
		shape=group.entities.add_instance(kicad_shape3d[@id],pos)
		shape.transform! mirror
		shape.transform! rotate
	end
		
	
end

class Drawing < Kicad_base

	attr_accessor :type, :Xstart, :Ystart, :Xend, :Yend, :Xcentre, :Ycentre, :angle, :width, :layer
	
	UKNOWN_TYPE=0
	SEGMENT=1
	CIRCLE=2
	ARC=3
	
	def initialize
		@type=UKNOWN_TYPE
		@Xstart=0
		@Ystart=0
		@Xend=0
		@Yend=0
		@Xcentre=0
		@Ycentre=0
		@angle=0
		@width=0
		@layer=0
	end
	
	def draw(group,m_point,r_angle,comp_layer)
		kicad_model = Sketchup.active_model
		kicad_layers = Sketchup.active_model.layers
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[@layer]
		end
		
		#rotate the silk component
		
		if(comp_layer==COPPER_TOP)
			m_point.z=@@z[TOP_SILK_SCREEN]
			mirror=0
		else
			m_point.z=@@z[BOTTOM_SILK_SCREEN]
			mirror=1
		end
				

		case @type
			when SEGMENT
				start_line=Geom::Point3d.new(@Xstart*@@unit+m_point.x,@Ystart*@@unit+m_point.y,m_point.z)
				end_line=Geom::Point3d.new(@Xend*@@unit+m_point.x,@Yend*@@unit+m_point.y,m_point.z)			
				if(r_angle!=0)
					rotate(start_line,m_point,r_angle*0.1)
					rotate(end_line,m_point,r_angle*0.1)
				end
				group.entities.add_line start_line,end_line
				

			when CIRCLE
				center=Geom::Point3d.new(@Xcentre*@@unit+m_point.x,@Ycentre*@@unit+m_point.y,m_point.z)
				if(r_angle!=0)
					rotate(center,m_point,r_angle*0.1)
				end
				if(mirror==1)
					mirror_x(center,m_point)
				end
				vector=Geom::Vector3d.new(0,0,1)
				radius=(@Xcentre-@Xend)!=0 ? @Xcentre-@Xend : @Ycentre-@Yend
				if(radius<0) then radius*=-1
				end
				group.entities.add_circle center,vector,radius*@@unit
			when ARC
				arc_center=Geom::Point3d.new(@Xcentre*@@unit+m_point.x,@Ycentre*@@unit+m_point.y,m_point.z)
				e_point=Geom::Point3d.new(@Xend*@@unit+m_point.x,@Yend*@@unit+m_point.y,m_point.z)
				if(r_angle!=0)
					rotate(arc_center,m_point,r_angle*0.1)
				end
				if(mirror==1)
					mirror_x(arc_center,m_point)
				end
				vector=arc_center.vector_to(e_point)
				vector.normalize!
				normal=Geom::Vector3d.new(0,0,-1)
				radius=(@Xcentre-@Xend)!=0 ? @Xcentre-@Xend : @Ycentre-@Yend
				if(radius<0) then radius*=-1
				end
				if(radius<0.20) then radius=0.20
				end
				s_angle=r_angle
				e_angle=s_angle+angle
				print("angle ",s_angle*0.1,"-",e_angle*0.1,"\n")
				arc=group.entities.add_arc arc_center,vector,normal,radius*@@unit,s_angle.degrees*0.1,e_angle.degrees*0.1
				arc
		end
	end
end

class Kicad_Module < Kicad_base

	attr_accessor :Xpos, :Ypos, :rotation, :layer, :drawing, :name, :reference
	attr_accessor :segment, :circle, :arc, :pad, :shape3d

	X=0
	Y=1
	Z=2

	STD=0
	SMD=1
	CONN=2
	HOLE=3
	MECA=4
	UKNOWN=5
			
	def initialize
		@Xpos=0
		@Ypos=0
		@rotation=0
		@layer=0
		@drawing=[]
		@name=""
		@reference=nil
		@pad=[]
		@shape3d=nil
			
	end
	
	def last_segment
		@segment.last
	end
	
	def last_circle
		@circle.last
	end
	
	def last_arc
		@circle.last
	end
	
	def last_pad
		@pad.last
	end
	
	def draw(group)
		
		kicad_model = Sketchup.active_model
		kicad_layers = Sketchup.active_model.layers
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[@layer]
		end
		
		module_pos=Geom::Point3d.new(@Xpos*@@unit,@Ypos*@@unit,0)
		
		
		#define main group
#		m_group=kicad_model.entities.add_group
		m_group=group.entities.add_group
		m_group.name=@name
		
		#define the silk group
		s_group=m_group.entities.add_group
		s_group.name="SILK"
				
		#Draw the pads
		
		#create the silk component
		if($kicad_settings.draw_silk==1)
			start_time=Time.now
			@drawing.each do |element|
				element.draw(s_group,module_pos,@rotation,@layer)
			end
			
			ref_pos=Geom::Point3d.new(module_pos.x+@reference.Xoff*@@unit,module_pos.y-@reference.Yoff*@@unit,0)
			rotations=@reference.Orientation/900
			if(rotations%2==0) then angle=0
			else angle=900
			end

			if(@layer==COPPER_TOP)
				ref_pos.z=@@z[TOP_SILK_SCREEN]
				mirror=0
			else
				ref_pos.z=@@z[BOTTOM_SILK_SCREEN]
				mirror=1
			end
			rotate(ref_pos,module_pos,@rotation*0.1)
			ref_text=draw_text(s_group,ref_pos,@reference.Xsize,@reference.Ysize,angle,mirror,@reference.Text)
			elaps_time=Time.now-start_time
			$report.update_drawing_elaps(elaps_time)
		end

		if($kicad_settings.draw_pad==1)
			start_time=Time.now
			@pad.each do |curr_pad|
			
				#define group for PAD

				case curr_pad.type
					
				when STD
					if (curr_pad.layer_mask & Pad::TOP)!=0
						print(curr_pad.layer_mask," & ",Pad::TOP," = ",curr_pad.layer_mask & Pad::TOP,"\n")
						module_pos.z=@@z[TOP_SILK_SCREEN]
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end
					if (curr_pad.layer_mask & Pad::BOTTOM)!=0
						print(curr_pad.layer_mask," & ",Pad::BOTTOM," = ",curr_pad.layer_mask & Pad::BOTTOM,"\n")
						module_pos.z=@@z[BOTTOM_SILK_SCREEN]
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end
				when SMD
					if (curr_pad.layer_mask & Pad::TOP)!=0
						print(curr_pad.layer_mask," & ",Pad::TOP," = ",curr_pad.layer_mask & Pad::TOP,"\n")
						module_pos.z=@@z[TOP_SILK_SCREEN]
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end
					if (curr_pad.layer_mask & Pad::BOTTOM)!=0
						module_pos.z=@@z[BOTTOM_SILK_SCREEN]
						print(curr_pad.layer_mask," & ",Pad::BOTTOM," = ",curr_pad.layer_mask & Pad::BOTTOM,"\n")
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end
				when CONN
					#nothing to do at the moment
					if (curr_pad.layer_mask & Pad::TOP)!=0
						print(curr_pad.layer_mask," & ",Pad::TOP," = ",curr_pad.layer_mask & Pad::TOP,"\n")
						module_pos.z=@@z[TOP_SILK_SCREEN]
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end
					if (curr_pad.layer_mask & Pad::BOTTOM)!=0
						module_pos.z=@@z[BOTTOM_SILK_SCREEN]
						print(curr_pad.layer_mask," & ",Pad::BOTTOM," = ",curr_pad.layer_mask & Pad::BOTTOM,"\n")
						curr_pad.draw(s_group,module_pos,@rotation,@layer)
					end

				when HOLE
					print("HOLE pad not implemented yet\n")
					#nothing to do at the moment

				when MECA
					#nothing to do at the moment
					print("MECA pad not implemented yet\n")
				end
				
				$report.inc_total_pads
			end
			elaps_time=Time.now-start_time
			$report.update_pad_elaps(elaps_time)
		end
		
		if($kicad_settings.draw_shape3d==1)
			if(@shape3d!=nil)
				if(@layer==COPPER_TOP) then module_pos.z=@@z[TOP_SILK_SCREEN]
				else module_pos.z=@@z[BOTTOM_SILK_SCREEN]
				end
				if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[SHAPE3D_LAYER]
				end
				@shape3d.place(m_group,module_pos,@rotation,@layer)
				$report.inc_total_shape3d
			else $report.inc_missing_shape3d
			end
		end
	end
			
end

class Board < Kicad_base

	COPPER_THICKNESS = 0.0039 # 0.1mm thickness for each layer
		
	attr_accessor :file_vers, :sheet_format, :sheet_size, :title, :date, :rev, :copper_layers
	attr_accessor :drawsegment, :text
	attr_reader :k_module

	def initialize
		@file_vers=0
		@boundings=[]
		@sheet_format=""
		@sheet_size=[]
		@title=""
		@date=""
		@rev=""
		@@unit=0.0001
		@copper_layers=16
		@k_module=[]
		@drawsegment=[]
		@mire=[]
		@track=[]
		@text=[]
		@thickness=$kicad_settings.thickness
	end
	
	def unit
		@@unit
	end
	
	def layer_name
		@@layer_name
	end
		
	def last_module
		return @k_module.last
	end
	
	def last_drawsegment
		@drawsegment.last
	end
	
	def last_mire
		@mire.last
	end
	
	def last_track
		@track.last
	end
	
	def draw

		start_time=Time.now
		#the following coordinates are used to place the pcb in a selected position
		
		@kicad_point=Geom::Point3d.new(0,0,0)

		Sketchup.status_text="Building layers"
		
		#calculate the positions of the bottom and top layers

		@@z[COPPER_TOP]=@thickness/2
		
		@@z[COPPER_BOTTOM]=-@thickness/2
		@@z[TOP_SILK_SCREEN]=@@z[COPPER_TOP]+@@unit
		@@z[BOTTOM_SILK_SCREEN]=@@z[COPPER_BOTTOM]-@@unit
		@@z[TOP_SIDE_ADHESIVE]=@@z[COPPER_TOP]+@@unit
		@@z[BOTTOM_SIDE_ADHESIVE]=@@z[COPPER_BOTTOM]-@@unit
		@@z[TOP_SOLDER_PASTE]=@@z[COPPER_TOP]+@@unit
		@@z[BOTTOM_SOLDER_PASTE]=@@z[COPPER_BOTTOM]-@@unit
		@@z[TOP_SOLDER_MASK]=@@z[COPPER_TOP]+@@unit
		@@z[BOTTOM_SOLDER_MASK]=@@z[COPPER_BOTTOM]-@@unit

		@@z[DRAWINGS]=@@z[COPPER_TOP]+@@unit
		@@z[COMMENT]=@@z[COPPER_TOP]+@@unit
		@@z[ECO1]=@@z[COPPER_TOP]+@@unit
		@@z[ECO2]=@@z[COPPER_TOP]+@@unit

		@@z[EDGE]=@thickness/2

		kicad_model=Sketchup.active_model
		kicad_layers = Sketchup.active_model.layers
		kicad_shape3d = kicad_model.definitions
		
		#create layers
		if($kicad_settings.multilayer==1) 
			@@layer_name.each_index do |index|
				if (index==0) then next
				end
				kicad_layers.add @@layer_name[index]
			end
		else
			kicad_layers.add "pcb"
			kicad_model.active_layer=kicad_layers.at("pcb")
		end
		
		#build Component definitions for SHAPE3D

		Sketchup.status_text="Catching components"

		inventory_shape3ds()

	
		#define PCB groups
		
		pcb_group=kicad_model.active_entities.add_group
		pcb_group.name="PCB"
		no_pcb_group=kicad_model.active_entities.add_group
		no_pcb_group.name="NO_PCB"
		
	
		#create pcb substrate with layer EDGE
		
		Sketchup.status_text="Drawing PCB substrate"

		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[EDGE]
		end
		
		pcb_edges=[]
		@drawsegment.each do |segment|
			if(segment.layer==EDGE) then pcb_edges+=segment.draw(pcb_group)
			else segment.draw(no_pcb_group)
			end
		end
		
		face=pcb_group.entities.add_face(pcb_edges)
		if(!face)
			UI.messagebox("Can't build pcb substrate")
			return 0
		end
		
		face.material=$kicad_settings.pcb_color
		face.back_material=$kicad_settings.pcb_color
		vector=face.normal
		vector.normalize
		if(vector.z==-1) then face.reverse!
		end
		face.pushpull -@thickness

		Sketchup.status_text="Drawing PCB text"

		@text.each do |item|
			group=pcb_group.entities.add_group
#			print(item.text,"\n")
			item.draw(group)
		end

		#Draw Modules
		@k_module.each do |curr_module|
			if(curr_module.name!="SKP")
				Sketchup.status_text="Drawing #{curr_module.reference.Text}"
				$report.inc_total_components
				curr_module.draw(pcb_group)
			else
				xpos=curr_module.Xpos*@@unit
				ypos=curr_module.Ypos*@@unit
				if(curr_module.layer==COPPER_TOP) then zpos=@@z[COPPER_TOP]
				else zpos=@@z[COPPER_BOTTOM]
				end
				@kicad_point=Geom::Point3d.new(xpos,ypos,zpos)
#				print("1st kicad point at ",@kicad_point.x," ",@kicad_point.y," ",@kicad_point.z,"\n")
			end
		end

		
		#Drill the PCB
		if($kicad_settings.drill_pcb==1)
			Sketchup.status_text="Drilling PCB substrate"
			drill_pcb(pcb_group)
		end

		kicad_layers.purge_unused
		
		#Autoplacement procedure

		
		if($kicad_settings.autoplace==1)
			Sketchup.status_text="Autoplacement"
			kicad_model.definitions.each do |definition|
				if(definition.name=="kicad")
					kicad_definition=definition
					kicad_model.entities.each do |object|
						if(object.typename=="ComponentInstance")
							if(object.definition==kicad_definition)
								orientations=object.transformation
								pcb_group.transform! orientations
								@kicad_point.transform! orientations
								kicad_box=object.bounds
								kicad_center=kicad_box.center
								# find the MODULE SKP in Sketchup and save the center of the bounding box
#								print("2nd kicad point at ",@kicad_point.x," ",@kicad_point.y," ",@kicad_point.z,"\n")
#								print("kicad center at ",kicad_center.x," ",kicad_center.y," ",kicad_center.z,"\n")
								displacement=@kicad_point.vector_to kicad_center
								pcb_group.transform! displacement
								break
							end
						end
					end
				end
			end
		end
		
		elaps_time=Time.now-start_time
		$report.update_total_elaps(elaps_time)

		$report.pcb_title=@title
		$report.pcb_rev=@rev
		$report.pcb_date=@date

	end
	

	def inventory_faces(group)
		faces=[]
		group.entities.each {|ent|
		faces.push(ent) if ent.is_a? Sketchup::Face }
		return faces
	end

	def inventory_shape3ds()
	
		kicad_model=Sketchup.active_model
		kicad_shape3d = kicad_model.definitions
	
		@k_module.each do |curr_module|
			count=0
			if(curr_module.shape3d==nil)
				next
			end
			kicad_shape3d.each do |curr_shape|
				if(curr_shape.name==curr_module.shape3d.filename)
					curr_module.shape3d.id=count
					break
				end
				count+=1
			end
			if(kicad_shape3d.count==count)
				if($kicad_settings.sys_location==1)
					shape_path=Sketchup.find_support_file curr_module.shape3d.filename+".skp", "Components/Kicad/"
				else shape_path=$kicad_settings.user_path+"/"+curr_module.shape3d.filename+".skp"
				end
				if(shape_path)
					if(File.exist?(shape_path))
						kicad_shape3d.load shape_path
						curr_module.shape3d.id=kicad_shape3d.count-1
					else curr_module.shape3d=nil #if no SHAPE3D was found for the current k_module the variable is set to nil
					end
				else curr_module.shape3d=nil
				end
			end
		end
	
	end

	def drill_pcb(group)

		start_time=Time.now
		faces=inventory_faces(group)

		kicad_model=Sketchup.active_model
		kicad_layers = Sketchup.active_model.layers
		if($kicad_settings.multilayer==1) then kicad_model.active_layer=kicad_layers[EDGE]
		end

		#Create drills
		module_position=Geom::Point3d.new(0,0,@@z[EDGE])
		pad_position=Geom::Point3d.new(0,0,0)
		
		@k_module.each do |curr_module|
			module_position.x=curr_module.Xpos*@@unit
			module_position.y=curr_module.Ypos*@@unit
			
			curr_module.pad.each do |curr_pad|
				if(curr_pad.drill!=nil) 
					pad_position.x=curr_pad.Xpos*@@unit+module_position.x
					pad_position.y=curr_pad.Ypos*@@unit+module_position.y
					pad_position.z=module_position.z
					if(curr_module.rotation.degrees*0.1!=0)
						rotate(pad_position,module_position,curr_module.rotation*0.1)
					end
					case curr_pad.drill.type
						when Kicad_drill::CIRCLE
#							print("drawing circle drill\n")
							center=curr_pad.drill.center(pad_position)
							if(curr_pad.drill.Xdrill>=0.020)
#								print("\tpad position ",pad_position.x," ",pad_position.y," ",pad_position.z,"\n")
								drill_circle=draw_circle(group,center,curr_pad.drill.Xdrill)
								group.entities.add_face(drill_circle)
							else $report.inc_missing_drills
							end
						when Kicad_drill::OBLONG
#							print("drawing oblong drill\n")
							center=curr_pad.drill.center(pad_position)
							if((curr_pad.drill.Xdrill>=0.020) && (curr_pad.drill.Ydrill>=0.020))
								drill_oblong=draw_oblong(group,center,curr_pad.drill.Xdrill,curr_pad.drill.Ydrill,curr_pad.Orientation*0.1)
								group.entities.add_face(drill_oblong)
							else $report.inc_missing_drills
							end

					end
#				else $report.inc_missing_drills
				end
			end
		end
		faces=inventory_faces(group)-faces
		faces.each do |face|
			face.pushpull -$kicad_settings.thickness()
		end
		elaps_time=Time.now-start_time
		$report.update_drill_elaps(elaps_time)
		
	end
	
end
